# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Badge to display information."""

from gi.repository import Adw, Gtk


@Gtk.Template(resource_path="/fr/romainvigier/MetadataCleaner/ui/Badge.ui")
class Badge(Adw.Bin):
    """Badge to display information."""

    __gtype_name__ = "Badge"



#!/usr/bin/env python3

# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Get release notes for a specific release."""


import argparse
import re

from typing import Dict, List


releases: Dict[str, str] = {}


def parse_changelog(path: str):
    """Parse a changelog file for versions and release notes.

    Args:
        path (str): Path to the changelog file.
    """
    release_regex = re.compile(r"^##\s")
    version_regex = re.compile(r"([\d\.]+)")

    current_version = ''
    current_notes: List[str] = []

    with open(path, "r") as f:
        for line in f.readlines():
            if release_regex.match(line):
                if current_version != '':
                    releases[current_version] = ''.join(current_notes).strip()
                current_notes = []
                current_version = version_regex.findall(line)[0]
            else:
                current_notes.append(line)
        releases[current_version] = ''.join(current_notes).strip()


def print_release_notes(version: str):
    """Print the release notes of the given release.

    Args:
        version (str): The requested version.
    """
    print(releases[version.replace("v", "")])


argparser = argparse.ArgumentParser()
argparser.add_argument(
    "-f", "--file",
    default="CHANGELOG.md",
    help="Path to the changelog file",
    metavar="CHANGELOG")
argparser.add_argument("version", help="Version to get the release notes for")
args = argparser.parse_args()


parse_changelog(args.file)
print_release_notes(args.version)



#!/usr/bin/env python3

# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Entry script for Metadata Cleaner."""

import gettext
import gi
import locale
import logging
import mimetypes
import os
import signal
import sys

gi.require_version("Gdk", "4.0")
gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")


APP_ID = "@app_id@"
DEVEL = "@devel@" == "True"
VERSION = "@version@"
LOCALE_DIR = "@localedir@"
PKGDATA_DIR = "@pkgdatadir@"
PKGSRC_DIR = "@pkgsrcdir@"

sys.path.insert(1, PKGSRC_DIR)


def setup_i18n() -> None:
    """Set the text domain for translations."""
    try:
        locale.bindtextdomain(APP_ID, LOCALE_DIR)  # type: ignore
        locale.textdomain(APP_ID)  # type: ignore
    except AttributeError as e:
        logging.warning(
            f"Unable to set the gettext translation domain.\nError:\n{e}")
    gettext.bindtextdomain(APP_ID, LOCALE_DIR)
    gettext.textdomain(APP_ID)


def setup_resources() -> None:
    """Load the application resources."""
    from gi.repository import Gio
    resource = Gio.Resource.load(
        os.path.join(PKGDATA_DIR, f"{APP_ID}.gresource"))
    Gio.Resource._register(resource)


def setup_mimetypes() -> None:
    """Add a mime.types file for the Flatpak runtime."""
    if os.path.exists("/.flatpak-info"):
        mimetypes.init(files=mimetypes.knownfiles + ["/app/share/mime.types"])


def run_app() -> int:
    """Run the application.

    Returns:
        int: Exit status
    """
    from metadatacleaner.app import MetadataCleaner
    app = MetadataCleaner(application_id=APP_ID, devel=DEVEL, version=VERSION)
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    exit_status = app.run(sys.argv)
    return exit_status


if __name__ == "__main__":
    setup_i18n()
    setup_resources()
    setup_mimetypes()

    exit_status = run_app()
    sys.exit(exit_status)



#!/usr/bin/env python3

# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Prepare the Flatpak manifest for a release build."""

import yaml


MANIFEST_FILE = "build-aux/fr.romainvigier.MetadataCleaner.yaml"


def rewrite_manifest() -> None:
    """Rewrite manifest to remove the devel flag."""
    print("Rewriting manifest for release...")

    with open(MANIFEST_FILE, "r") as f:
        manifest = yaml.safe_load(f)

    manifest["default-branch"] = "stable"

    try:
        opts = manifest["modules"][-1]["config-opts"]
    except KeyError:
        return
    opts = ["-Ddevel=false" if i == "-Ddevel=true" else i for i in opts]
    manifest["modules"][-1]["config-opts"] = opts

    with open(MANIFEST_FILE, "w") as f:
        yaml.dump(manifest, f)

    print("Finished rewriting manifest.")


rewrite_manifest()



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Dialog warning the user of possible data loss on cleaning."""

from gi.repository import Gio, GObject, Gtk


@Gtk.Template(
    resource_path=(
        "/fr/romainvigier/MetadataCleaner/ui/CleaningWarningDialog.ui")
)
class CleaningWarningDialog(Gtk.MessageDialog):
    """Dialog warning the user of possible data loss on cleaning."""

    __gtype_name__ = "CleaningWarningDialog"

    settings = GObject.Property(type=Gio.Settings)

    _checkbutton: Gtk.CheckButton = Gtk.Template.Child()

    @Gtk.Template.Callback()
    def _on_settings_changed(self, dialog, p_spec: GObject.ParamSpec) -> None:
        if not self.settings:
            return
        self.settings.bind(
            "cleaning-without-warning",
            self._checkbutton,
            "active",
            Gio.SettingsBindFlags.DEFAULT)



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Indicator showing the status of the files manager."""

from gettext import gettext as _
from gettext import ngettext as __
from gi.repository import Gio, GLib, GObject, Gtk

from metadatacleaner.modules.filestore \
    import FileStore, FileStoreAction, FileStoreState


@Gtk.Template(
    resource_path="/fr/romainvigier/MetadataCleaner/ui/StatusIndicator.ui"
)
class StatusIndicator(Gtk.Stack):
    """Indicator showing the status of the files manager."""

    __gtype_name__ = "StatusIndicator"

    file_store: FileStore = GObject.Property(type=FileStore, nick="file-store")

    _progressbar: Gtk.ProgressBar = Gtk.Template.Child()
    _done_label: Gtk.Label = Gtk.Template.Child()

    def __init__(self, *args, **kwargs) -> None:
        """Status indicator initialization."""
        super().__init__(*args, **kwargs)
        self.show_idle()

    def _sync_progressbar(self, current, total) -> None:
        if not self.file_store or not self.file_store.last_action:
            return
        text = {
            FileStoreAction.ADDING:
                _("Processing file {}/{}").format(current, total),
            FileStoreAction.CLEANING:
                _("Cleaning file {}/{}").format(current, total),
        }
        self._progressbar.set_text(text[self.file_store.last_action])
        self._progressbar.set_fraction(current / total if total > 0 else 0)

    @Gtk.Template.Callback()
    def _on_file_store_changed(
            self,
            widget: Gtk.Widget,
            pspec: GObject.ParamSpec) -> None:
        self.file_store.connect(
            "state-changed",
            self._on_file_store_state_changed)
        self.file_store.connect(
            "progress-changed",
            self._on_file_store_progress_changed)

    @Gtk.Template.Callback()
    def _on_cancel_button_clicked(self, button: Gtk.Button) -> None:
        if self.file_store.last_action == FileStoreAction.ADDING:
            self.file_store.cancel_addding_gfiles()
        elif self.file_store.last_action == FileStoreAction.CLEANING:
            self.file_store.cancel_cleaning_files()

    def _on_file_store_state_changed(
            self,
            file_store: FileStore,
            new_state: FileStoreState) -> None:
        if new_state == FileStoreState.WORKING:
            self.show_progressbar()

    def _on_file_store_progress_changed(
            self,
            file_store: FileStore,
            current: int,
            total: int) -> None:
        self._sync_progressbar(current, total)
        if current == total:
            if file_store.last_action == FileStoreAction.CLEANING:
                clean_message = __(
                    "%i file cleaned.",
                    "%i files cleaned.",
                    len(file_store.get_cleaned_files())
                ) % len(file_store.get_cleaned_files())
                error_message = (__(
                    "%i error occured.",
                    "%i errors occured.",
                    len(file_store.get_errored_files())
                ) % len(file_store.get_errored_files())
                    if len(file_store.get_errored_files()) > 0
                    else "")
                self._done_label.set_label(
                    " ".join([clean_message, error_message]))
                if not self.get_root().is_active():
                    self.send_done_notification()
            else:
                self._done_label.set_label("")
            self.show_done()

    def show_idle(self) -> None:
        """Show the idle state."""
        self.set_visible_child_name("idle")

    def show_progressbar(self) -> None:
        """Show a progress bar."""
        self.set_visible_child_name("working")

    def show_done(self) -> None:
        """Show a "Done" message."""
        self.set_visible_child_name("done")

    def send_done_notification(self) -> None:
        """Send a notification about the finished cleaning process."""
        window = self.get_root()
        app = window.get_application()
        notification = Gio.Notification.new(app.name)
        notification.set_body(self._done_label.get_label())
        notification.set_default_action_and_target(
            "app.show-window",
            GLib.Variant.new_uint32(window.get_id()))
        app.send_notification(f"done{window.get_id()}", notification)



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Folder chooser dialog."""

from gettext import gettext as _
from gi.repository import Gtk


class FolderChooserDialog(Gtk.FileChooserNative):
    """Folder chooser dialog."""

    __gtype_name__ = "FolderChooserDialog"

    def __init__(self, *args, **kwargs) -> None:
        """Folder chooser dialog initialization."""
        super().__init__(*args, **kwargs)
        self._setup_choice()

    def _setup_choice(self) -> None:
        self.add_choice(
            "recursive", _("Add files from subfolders"), None, None)
        self.set_choice("recursive", "true")



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Menu button."""

from gi.repository import Gtk


@Gtk.Template(
    resource_path="/fr/romainvigier/MetadataCleaner/ui/MenuButton.ui"
)
class MenuButton(Gtk.MenuButton):
    """Menu button."""

    __gtype_name__ = "MenuButton"



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""View details of a file."""

from gettext import gettext as _
from gi.repository import Adw, GLib, Gtk
from typing import Optional

from metadatacleaner.modules.file import File, FileState

from metadatacleaner.ui.metadataview import MetadataView


@Gtk.Template(
    resource_path="/fr/romainvigier/MetadataCleaner/ui/DetailsView.ui"
)
class DetailsView(Gtk.Widget):
    """View details of a file."""

    __gtype_name__ = "DetailsView"

    _child: Optional[Gtk.Widget] = None

    def view_file(self, f: File) -> None:
        """Set the file to view.

        Args:
            f (File): The file to view.
        """
        self.clear()
        if f.state == FileState.HAS_METADATA:
            self._setup_metadata_details(f)
        elif f.state == FileState.CLEANED:
            self._setup_cleaned_details()
        elif f.state in [
                FileState.ERROR_WHILE_CHECKING_METADATA,
                FileState.ERROR_WHILE_INITIALIZING,
                FileState.ERROR_WHILE_REMOVING_METADATA,
                FileState.HAS_NO_METADATA,
                FileState.UNSUPPORTED]:
            self._setup_error_details(f)

    def set_child(self, w) -> None:
        """Set the widget's child."""
        self.clear()
        w.set_parent(self)
        self._child = w

    def clear(self) -> None:
        """Clear the view."""
        if self._child:
            self._child.unparent()

    def _setup_cleaned_details(self) -> None:
        self.set_child(
            Adw.StatusPage(
                title=_("The File Has Been Cleaned"),
                description=_(
                    "Known metadata have been removed, however the cleaning "
                    "process has some limitations."),
                child=Gtk.Button(
                    label=_("Learn more"),
                    action_name="app.help",
                    action_target=GLib.Variant.new_string(
                        "/general#cleaning-process"),
                    halign=Gtk.Align.CENTER,
                    css_classes=["pill", "suggested-action"]),
                css_classes=["compact"]))

    def _setup_metadata_details(self, f: File) -> None:
        self.set_child(MetadataView(metadata=f.metadata))

    def _setup_error_details(self, f: File) -> None:
        info_titles = {
            FileState.ERROR_WHILE_INITIALIZING: _("Unable to Read the File"),
            FileState.UNSUPPORTED: _("File Type not Supported"),
            FileState.ERROR_WHILE_CHECKING_METADATA: _(
                "Unable to Check for Metadata"),
            FileState.HAS_NO_METADATA: _("No Known Metadata"),
            FileState.ERROR_WHILE_REMOVING_METADATA: _(
                "Unable to Remove Metadata")
        }
        info_details = str(f.error or "")
        if f.state == FileState.HAS_NO_METADATA:
            info_details = _("The file will be cleaned anyway to be sure.")
        self.set_child(
            Adw.StatusPage(
                title=info_titles[f.state],
                description=info_details,
                css_classes=["compact"]))



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Application window of Metadata Cleaner."""

from gettext import gettext as _
from gi.repository import Adw, Gdk, Gio, GLib, GObject, Gtk
from typing import Any

from metadatacleaner.modules.filestore import FileStore, FileStoreState

from metadatacleaner.ui.addfilesbutton import AddFilesButton
from metadatacleaner.ui.emptyview import EmptyView
from metadatacleaner.ui.filechooserdialog import FileChooserDialog
from metadatacleaner.ui.filesview import FilesView
from metadatacleaner.ui.folderchooserdialog import FolderChooserDialog
from metadatacleaner.ui.menubutton import MenuButton
from metadatacleaner.ui.detailsview import DetailsView
from metadatacleaner.ui.cleaningwarningdialog import CleaningWarningDialog


@Gtk.Template(resource_path="/fr/romainvigier/MetadataCleaner/ui/Window.ui")
class Window(Adw.ApplicationWindow):
    """Application window of Metadata Cleaner."""

    __gtype_name__ = "Window"

    file_store = GObject.Property(type=FileStore, nick="file-store")

    _mode_flap: Adw.Flap = Gtk.Template.Child()
    _view_stack: Gtk.Stack = Gtk.Template.Child()
    _details_view: DetailsView = Gtk.Template.Child()

    _about_window: Adw.AboutWindow = Gtk.Template.Child()
    _file_chooser_dialog: FileChooserDialog = Gtk.Template.Child()
    _folder_chooser_dialog: FolderChooserDialog = Gtk.Template.Child()
    _cleaning_warning_dialog: CleaningWarningDialog = Gtk.Template.Child()

    def __init__(
        self,
        *args,
        **kwargs
    ) -> None:
        """Window initialization."""
        app = kwargs["application"]
        super().__init__(
            title=app.name,
            *args,
            **kwargs
        )
        self._setup_size()
        self._setup_devel_style()
        self._setup_file_store()
        self._setup_about_window()
        self._setup_drop_target()
        self._setup_actions()

    # SETUP #

    def _setup_size(self) -> None:
        def on_close_request(window: Gtk.Window) -> None:
            width, height = self.get_default_size()
            maximized = self.is_maximized()
            self.get_application().settings.set_uint("window-width", width)
            self.get_application().settings.set_uint("window-height", height)
            self.get_application().settings.set_boolean(
                "window-maximized", maximized)
        self.set_default_size(
            self.get_application().settings.get_uint("window-width"),
            self.get_application().settings.get_uint("window-height"))
        if self.get_application().settings.get_boolean("window-maximized"):
            self.maximize()
        self.connect("close-request", on_close_request)

    def _setup_devel_style(self) -> None:
        if self.get_application().devel:
            self.add_css_class("devel")

    def _setup_file_store(self) -> None:

        def on_items_changed(
                file_store: FileStore,
                position: int,
                added: int,
                removed: int) -> None:
            if len(file_store) == 0:
                self.lookup_action("clean-metadata").set_enabled(False)
                self.close_details_view()
                self.show_empty_view()
            else:
                self.show_files_view()

        def on_state_changed(
                file_store: FileStore,
                file_index: int) -> None:
            self.lookup_action("clean-metadata").set_enabled(not (
                file_store.state == FileStoreState.WORKING
                or len(file_store.get_cleanable_files()) == 0))

        self.file_store = FileStore()
        self.file_store.connect("items-changed", on_items_changed)
        self.file_store.connect("file-state-changed", on_state_changed)
        self.file_store.connect("state-changed", on_state_changed)
        self.get_application().settings.bind(
            "lightweight-cleaning",
            self.file_store,
            "lightweight-mode",
            Gio.SettingsBindFlags.DEFAULT)

    def _setup_about_window(self) -> None:
        self._about_window.add_acknowledgement_section(
            _("Libraries"), ["mat2 https://0xacab.org/jvoisin/mat2"])

    def _setup_drop_target(self) -> None:

        def on_drop(
                widget: Gtk.DropTarget,
                value: Any,
                x: int,
                y: int):
            if isinstance(value, Gdk.FileList):
                self.file_store.add_gfiles(value.get_files())

        drop_target = Gtk.DropTarget.new(Gdk.FileList, Gdk.DragAction.COPY)
        drop_target.connect("drop", on_drop)
        self.add_controller(drop_target)

    def _setup_actions(self) -> None:

        def on_close(action: Gio.Action, parameters: None) -> None:
            self.destroy()
        close = Gio.SimpleAction.new("close", None)
        close.connect("activate", on_close)
        self.add_action(close)

        def on_about(action: Gio.Action, parameters: None) -> None:
            self._about_window.show()
        about = Gio.SimpleAction.new("about", None)
        about.connect("activate", on_about)
        self.add_action(about)

        def on_add_files(action: Gio.Action, parameters: None) -> None:
            self._file_chooser_dialog.show()
        add_files = Gio.SimpleAction.new("add-files", None)
        add_files.connect("activate", on_add_files)
        self.add_action(add_files)

        def on_add_folders(action: Gio.Action, parameters: None) -> None:
            self._folder_chooser_dialog.show()
        add_folders = Gio.SimpleAction.new("add-folders", None)
        add_folders.connect("activate", on_add_folders)
        self.add_action(add_folders)

        def on_remove_file(
                action: Gio.Action,
                parameters: GLib.Variant) -> None:
            self.file_store.remove_file_with_index(parameters.get_uint32())
        remove_file = Gio.SimpleAction.new(
            "remove-file",
            GLib.VariantType.new("u"))
        remove_file.connect("activate", on_remove_file)
        self.add_action(remove_file)

        def on_clear_files(action: Gio.Action, parameters: None) -> None:
            self.file_store.remove_files()
        clear_files = Gio.SimpleAction.new("clear-files", None)
        clear_files.connect("activate", on_clear_files)
        clear_files.set_enabled(False)
        self.add_action(clear_files)

        def on_view_details(
                action: Gio.Action,
                parameters: GLib.Variant) -> None:
            f = self.file_store.get_file_with_index(parameters.get_uint32())
            self._details_view.view_file(f)
            self.open_details_view()
        view_details = Gio.SimpleAction.new(
            "view-details",
            GLib.VariantType.new("u"))
        view_details.connect("activate", on_view_details)
        self.add_action(view_details)

        def on_close_details_view(
                action: Gio.Action,
                parameters: None) -> None:
            self.close_details_view()
        close_details_view = Gio.SimpleAction.new(
            "close-details-view",
            None)
        close_details_view.connect("activate", on_close_details_view)
        self.add_action(close_details_view)

        def on_clean_metadata(action: Gio.Action, parameters: None) -> None:
            self.close_details_view()
            if not self.get_application() \
                    .settings.get_boolean("cleaning-without-warning"):
                self._cleaning_warning_dialog.show()
                return
            self.file_store.clean_files()
        clean_metadata = Gio.SimpleAction.new("clean-metadata", None)
        clean_metadata.connect("activate", on_clean_metadata)
        clean_metadata.set_enabled(False)
        self.add_action(clean_metadata)

    # SIGNALS

    @Gtk.Template.Callback()
    def _on_mode_flap_revealed(
            self,
            flap: Adw.Flap,
            p_spec: GObject.ParamSpec) -> None:
        if not flap.get_reveal_flap():
            self.close_details_view()

    @Gtk.Template.Callback()
    def _on_file_chooser_dialog_response(
            self,
            dialog: FileChooserDialog,
            response: Gtk.ResponseType) -> None:
        dialog.hide()
        if response == Gtk.ResponseType.ACCEPT:
            self.file_store.add_gfiles(dialog.get_files())

    @Gtk.Template.Callback()
    def _on_folder_chooser_dialog_response(
            self,
            dialog: FolderChooserDialog,
            response: Gtk.ResponseType) -> None:
        dialog.hide()
        if response == Gtk.ResponseType.ACCEPT:
            self.file_store.add_gfiles(
                dialog.get_files(),
                dialog.get_choice("recursive") == "true")

    @Gtk.Template.Callback()
    def _on_cleaning_warning_dialog_response(
            self,
            dialog: CleaningWarningDialog,
            response: Gtk.ResponseType) -> None:
        dialog.hide()
        if response == Gtk.ResponseType.OK:
            self.file_store.clean_files()

    # VIEWS #

    def show_empty_view(self) -> None:
        """Show an empty view."""
        self._view_stack.set_visible_child_name("empty")
        self.lookup_action("clear-files").set_enabled(False)

    def show_files_view(self) -> None:
        """Show the files."""
        self._view_stack.set_visible_child_name("files")
        self.lookup_action("clear-files").set_enabled(True)

    def open_details_view(self) -> None:
        """Show the details view."""
        self._mode_flap.set_reveal_flap(True)

    def close_details_view(self) -> None:
        """Close the details view."""
        self._mode_flap.set_reveal_flap(False)
        self._view_stack.get_child_by_name("files").clear_selected_file()



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""View when no files are added."""

from gi.repository import Gtk


@Gtk.Template(resource_path="/fr/romainvigier/MetadataCleaner/ui/EmptyView.ui")
class EmptyView(Gtk.Box):
    """View when no files are added."""

    __gtype_name__ = "EmptyView"



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""File chooser dialog."""

from gettext import gettext as _
from gi.repository import Gtk

from metadatacleaner.modules.filestore import SUPPORTED_FORMATS


class FileChooserDialog(Gtk.FileChooserNative):
    """File chooser dialog."""

    __gtype_name__ = "FileChooserDialog"

    def __init__(self, *args, **kwargs) -> None:
        """File chooser dialog initialization."""
        super().__init__(*args, **kwargs)
        self._setup_file_filter()

    def _setup_file_filter(self) -> None:
        file_filter = Gtk.FileFilter()
        file_filter.set_name(_("All supported files"))
        for mimetype, extensions in SUPPORTED_FORMATS.items():
            for extension in extensions:
                file_filter.add_suffix(extension[1:])
        self.add_filter(file_filter)



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Row displaying a metadata."""

from gi.repository import GObject, Gtk


@Gtk.Template(
    resource_path="/fr/romainvigier/MetadataCleaner/ui/MetadataDetailsRow.ui"
)
class MetadataDetailsRow(Gtk.ListBoxRow):
    """Row displaying a metadata."""

    __gtype_name__ = "MetadataDetailsRow"

    key = GObject.Property(type=str, default="")
    value = GObject.Property(type=str, default="")



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Files Manager object and states."""

import libmat2
import mimetypes

from concurrent.futures import ThreadPoolExecutor, as_completed
from enum import IntEnum, auto
from gi.repository import Gio, GLib, GObject
from threading import Thread
from typing import Dict, Iterable, List, Optional

from metadatacleaner.modules.file import File, FileState
from metadatacleaner.modules.logger import Logger as logger


def _get_supported_formats() -> Dict:
    formats = {}
    for parser in libmat2.parser_factory._get_parsers():
        for mimetype in parser.mimetypes:
            extensions = set()
            for extension in mimetypes.guess_all_extensions(mimetype):
                if extension not in libmat2.UNSUPPORTED_EXTENSIONS:
                    extensions.add(extension)
            if not extensions:
                continue
            formats[mimetype] = extensions
    return formats


SUPPORTED_FORMATS = _get_supported_formats()


class FileStoreState(IntEnum):
    """States the Files Manager can have."""

    IDLE = auto()
    WORKING = auto()


class FileStoreAction(IntEnum):
    """Actions the Files manager can do."""

    ADDING = auto()
    CLEANING = auto()


class FileStore(Gio.ListStore):
    """File Store object."""

    __gtype_name__ = "FileStore"

    __gsignals__ = {
        "file-state-changed": (GObject.SIGNAL_RUN_LAST, None, (int,)),
        "state-changed": (GObject.SIGNAL_RUN_LAST, None, (int,)),
        "progress-changed": (GObject.SIGNAL_RUN_LAST, None, (int, int))
    }

    lightweight_mode: bool = GObject.Property(
        type=bool,
        nick="lightweight-mode",
        default=False)

    def __init__(self) -> None:
        """File Store initialization."""
        Gio.ListStore.__init__(self, item_type=File)
        self.state = FileStoreState.IDLE
        self.last_action: Optional[FileStoreAction] = None
        self.progress = (0, 0)
        self.add_files_executor = ThreadPoolExecutor()
        self.clean_files_executor = ThreadPoolExecutor()

    def _on_file_state_changed(self, f: File, new_state: FileState) -> None:
        def emit() -> bool:
            self.emit("file-state-changed", self.get_index_of_file(f))
            return GLib.SOURCE_REMOVE
        GLib.idle_add(emit)

    def _set_state(self, state: FileStoreState) -> None:
        if state == self.state:
            return
        self.state = state

        def emit() -> bool:
            self.emit("state-changed", state)
            return GLib.SOURCE_REMOVE
        GLib.idle_add(emit)

    def _set_progress(self, current: int, total: int) -> None:
        self.progress = (current, total)

        def emit() -> bool:
            self.emit("progress-changed", current, total)
            return GLib.SOURCE_REMOVE
        GLib.idle_add(emit)

    def get_files(self) -> List[File]:
        """Get all the files from the File Store.

        Returns:
            List[File]: List of files.
        """
        return self

    def get_file_with_index(self, index: int) -> File:
        """Get a file at a specified index.

        Args:
            index (int): The desired index.

        Returns:
            Optional[File]: The file if found, else None.
        """
        f = self.get_item(index)
        if not f:
            raise RuntimeError("No file in the file store for this index.")
        return f

    def get_index_of_file(self, f: File) -> int:
        """Get the index of a given file.

        Args:
            f (File): The file to get the index for.

        Returns:
            int: The file index.
        """
        found, position = self.find(f)
        if not found:
            raise RuntimeError("File not found in file store.")
        return position

    def add_gfiles(
            self, gfiles: List[Gio.File], recursive: bool = True) -> None:
        """Add Gio Files to the Files Manager.

        Args:
            gfiles (List[Gio.File]): List of Gio Files to add.
            recursive (bool, optional): If subdirectories should also be looked
            into. Defaults to True.
        """
        thread = Thread(
            target=self._add_gfiles_async,
            args=(gfiles, recursive),
            daemon=True)
        thread.start()

    def _add_gfiles_async(
            self, gfiles: List[Gio.File], recursive: bool = True) -> None:
        self._set_state(FileStoreState.WORKING)
        all_gfiles = self._gather_all_gfiles(gfiles, recursive)
        self._set_progress(
            self.progress[0], self.progress[1] + len(all_gfiles))
        self.last_action = FileStoreAction.ADDING
        futures = {
            self.add_files_executor.submit(self._add_gfile, gfile)
            for gfile in all_gfiles
        }
        for future in as_completed(futures):
            current = self.progress[0] + 1
            total = self.progress[1]
            self._set_progress(current, total)
            if current == total:
                self._stop_adding_gfiles()

    def _gather_all_gfiles(
            self,
            gfiles: List[Gio.File],
            recursive: bool = True) -> List[Gio.File]:
        all_gfiles: List[Gio.File] = []
        for gfile in gfiles:
            if not gfile:
                continue
            f_type = gfile.query_file_type(Gio.FileQueryInfoFlags.NONE, None)
            if f_type == Gio.FileType.DIRECTORY:
                all_gfiles.extend(self._get_gfiles_from_dir(gfile, recursive))
            elif f_type == Gio.FileType.REGULAR:
                all_gfiles.append(gfile)
            else:
                logger.warning(
                    f"File {gfile.get_path()} is neither a directory nor a "
                    "regular file, skipping.")
        return all_gfiles

    def _get_gfiles_from_dir(
            self, dir: Gio.File, recursive: bool) -> List[Gio.File]:
        gfiles: List[Gio.File] = []
        subdirs: List[Gio.File] = []
        children_enumerator = dir.enumerate_children(
            "",
            Gio.FileQueryInfoFlags.NONE,
            None)
        while True:
            info = children_enumerator.next_file(None)
            if info is None:
                break
            child = children_enumerator.get_child(info)
            if info.get_file_type() == Gio.FileType.DIRECTORY:
                if recursive:
                    subdirs.append(child)
            elif info.get_file_type() == Gio.FileType.REGULAR:
                gfiles.append(child)
        children_enumerator.close(None)
        with ThreadPoolExecutor() as executor:
            for subgfiles in executor.map(
                    self._get_gfiles_from_dir,
                    subdirs,
                    [recursive] * len(subdirs)):
                gfiles.extend(subgfiles)
        return gfiles

    def _add_gfile(self, gfile: Gio.File) -> None:
        if not gfile.query_exists(None):
            logger.warning(
                f"File {gfile.get_path()} does not exist, skipping.")
            return

        if bool(list(filter(lambda x: x.path == gfile.get_path(), self))):
            logger.warning(f"Skipping {gfile.get_path()}, already added.")
            return

        f = File(gfile)
        f.check_metadata()

        def finish() -> bool:
            self.append(f)
            f.connect("state-changed", self._on_file_state_changed)
            return GLib.SOURCE_REMOVE
        GLib.idle_add(finish)

    def _stop_adding_gfiles(self) -> None:
        self.add_files_executor.shutdown(wait=False, cancel_futures=True)
        self.add_files_executor = ThreadPoolExecutor()
        self._set_state(FileStoreState.IDLE)
        self._set_progress(0, 0)

    def cancel_addding_gfiles(self) -> None:
        """Cancel adding GFiles."""
        self._stop_adding_gfiles()

    def remove_file(self, f: File) -> None:
        """Remove a file from the File Store.

        Args:
            f (File): The file to remove.
        """
        self.remove_file_with_index(self.get_index_of_file(f))

    def remove_file_with_index(self, index: int) -> None:
        """Remove the file with the given index from the File Store.

        Args:
            index (int): The index of the file to remove.
        """
        self.remove(index)

    def remove_files(self) -> None:
        """Remove all the files from the File Store."""
        self.remove_all()

    def clean_files(self) -> None:
        """Remove metadata from all the cleanable files."""
        thread = Thread(target=self._clean_files_async, daemon=True)
        thread.start()

    def _clean_files_async(self) -> None:
        cleanable_files = self.get_cleanable_files()
        self._set_progress(
            self.progress[0], self.progress[1] + len(cleanable_files))
        self._set_state(FileStoreState.WORKING)
        self.last_action = FileStoreAction.CLEANING
        futures = {
            self.clean_files_executor.submit(f.clean, self.lightweight_mode)
            for f in cleanable_files
        }
        for future in as_completed(futures):
            current = self.progress[0] + 1
            total = self.progress[1]
            self._set_progress(current, total)
        self._stop_cleaning_files()

    def _stop_cleaning_files(self) -> None:
        self.clean_files_executor.shutdown(wait=False, cancel_futures=True)
        self.clean_files_executor = ThreadPoolExecutor()
        self._set_state(FileStoreState.IDLE)
        self._set_progress(0, 0)

    def cancel_cleaning_files(self) -> None:
        """Cancel the cleaning process."""
        self._stop_cleaning_files()

    def get_cleanable_files(self) -> List[File]:
        """Get all the cleanable files.

        Returns:
            List[File]: List of cleanable files.
        """
        return self._get_files_with_states((
            FileState.HAS_METADATA,
            FileState.HAS_NO_METADATA))

    def get_cleaned_files(self) -> List[File]:
        """Get all the cleaned files.

        Returns:
            List[File]: List of cleaned files.
        """
        return self._get_files_with_states((FileState.CLEANED,))

    def get_errored_files(self) -> List[File]:
        """Get all the files with errors.

        Returns:
            List[File]: List of files with errors.
        """
        return self._get_files_with_states((
            FileState.ERROR_WHILE_INITIALIZING,
            FileState.ERROR_WHILE_CHECKING_METADATA,
            FileState.ERROR_WHILE_REMOVING_METADATA))

    def _get_files_with_states(
            self,
            states: Iterable[FileState]) -> List[File]:
        return list(filter(lambda f: f.state in states, self))



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Settings button."""

from gi.repository import Gtk


@Gtk.Template(
    resource_path="/fr/romainvigier/MetadataCleaner/ui/SettingsButton.ui"
)
class SettingsButton(Gtk.MenuButton):
    """Settings button."""

    __gtype_name__ = "SettingsButton"



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Button allowing to add files."""

from gi.repository import Adw, Gtk


@Gtk.Template(
    resource_path="/fr/romainvigier/MetadataCleaner/ui/AddFilesButton.ui"
)
class AddFilesButton(Gtk.Widget):
    """Button allowing to add files."""

    __gtype_name__ = "AddFilesButton"



#!/usr/bin/env python3

# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Run the screenshooter in a headless compositor."""

import os
import subprocess

from datetime import date
from langcodes import Language, LanguageTagError
from typing import Optional


PREFIX = os.path.join("/tmp", "screenshots")
DATA_DIR = os.path.join(PREFIX, "share")
LOCALE_DIR = os.path.join(DATA_DIR, "locale")

APP_ID = "fr.romainvigier.MetadataCleaner"
PACKAGE_NAME = "metadata-cleaner"
RESOURCE_DIR = os.path.join(DATA_DIR, PACKAGE_NAME)
RESOURCE_FILE = os.path.join(RESOURCE_DIR, f"{APP_ID}.gresource")
RESOURCE_PATH = "/fr/romainvigier/MetadataCleaner"
TEXTDOMAIN = APP_ID

SOCKET = "wayland-99"
SOCKET_2X = "wayland-100"


class Widget:
    """Base widget."""

    def __init__(
            self,
            ui_file: str,
            image_file: str,
            css_file: Optional[str] = None) -> None:
        """Create a new widget.

        Args:
            ui_file (str): Path to the UI file describing the widget.
            image_file (str): Path to the output image file.
            css_file (str, optional): Path to the stylesheet file to load for
                the widget. Defaults to None.
        """
        self.ui_file = ui_file
        self.image_file = image_file
        self.license_file = f"{image_file}.license"
        self.css_file: Optional[str] = css_file


class ApplicationWidget(Widget):
    """Application widget."""

    def __init__(self, number: int) -> None:
        """Create a new application widget.

        Args:
            number (int): Number of the screenshot.
        """
        ui_file = os.path.join("screenshots", f"application-{number}.ui")
        image_file = os.path.join("resources", "screenshots", f"{number}.png")
        css_file = os.path.join("screenshots", "application.css")
        super().__init__(ui_file, image_file, css_file)


class HelpWidget(Widget):
    """Help widget."""

    def __init__(self, name: str, langcode: str, css: bool) -> None:
        """Create a new help widget.

        Args:
            name (str): Name of the widget.
            lang (str): Language of the widget.
            css (bool): If the widget uses a custom stylesheet.
        """
        ui_file = os.path.join("screenshots", f"help-{name}.ui")
        image_file = os.path.join("help", langcode, "figures", f"{name}.png")
        css_file = os.path.join("screenshots", f"help-{name}.css") if css \
            else None
        super().__init__(ui_file, image_file, css_file)


def compile_translations() -> None:
    """Compile translations."""
    print("Compiling translations…")
    po_dir = os.path.join("application", "po")
    with open(os.path.join(po_dir, "LINGUAS")) as f:
        languages = [lang for lang in f.read().splitlines()
                     if lang != "" and lang[0] != "#"]
    languages.sort()
    for i, lang in enumerate(languages, 1):
        print(f"[{i}/{len(languages)}] Compiling {lang} translation…")
        in_file = os.path.join(po_dir, f"{lang}.po")
        out_dir = os.path.join(LOCALE_DIR, lang, "LC_MESSAGES")
        out_file = os.path.join(out_dir, f"{TEXTDOMAIN}.mo")
        os.makedirs(out_dir, exist_ok=True)
        subprocess.run([
            "msgfmt",
            f"--output-file={out_file}",
            in_file])


def compile_resources() -> None:
    """Compile resources."""
    print("Compiling resources…")
    source_dir = os.path.join("application", "data")
    source_file = os.path.join(source_dir, f"{APP_ID}.gresource.xml")
    os.makedirs(RESOURCE_DIR, exist_ok=True)
    subprocess.run([
        "glib-compile-resources",
        f"--target={RESOURCE_FILE}",
        f"--sourcedir={source_dir}",
        source_file])


def start_weston(
        scale: Optional[int] = None,
        socket: Optional[str] = None) -> subprocess.Popen[bytes]:
    """Start the Weston compositor in headless mode."""
    args = []
    if scale:
        args.append(f"--scale={scale}")
    if socket:
        args.append(f"--socket={socket}")
    return subprocess.Popen(
        ["weston", "--backend=headless-backend.so"] + args,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL)


def run_uishooter(
        ui_file: str,
        output: Optional[str] = None,
        resource_file: Optional[str] = None,
        resource_path: Optional[str] = None,
        textdomain: Optional[str] = None,
        locale: Optional[str] = None,
        locale_dir: Optional[str] = None,
        css: Optional[str] = None,
        scale: Optional[int] = None,
        dark: bool = False,
        libadwaita: bool = False) -> int:
    """Shoot the given UI file.

    Args:
        ui_file (str): Path to the UI file.
        output (str, optional): Path to the output image. Defaults to None.
        resource_file (str, optional): Path to a resource file to load.
            Defaults to None.
        resource_path (str, optional): Resource path to load. Defaults to None.
        textdomain (str, optional): Translation textdomain. Defaults to None.
        locale (str, optional): Locale to use. Defaults to None.
        locale_dir (str, optional): Path to a directory containing compiled
            translations. Defaults to None.
        css (str, optional): Path to a CSS file to load. Defaults to None.
        scale (int, optional): Integer scale factor of the output image.
            Defaults to 1.
        dark (bool, optional): Use dark color scheme. Defaults to False.
        libadwaita (bool, optional): Use libadwaita. Defaults to False.

    Returns:
        int: uishooter exit code
    """
    args = []
    env = os.environ.copy()
    if output:
        args.append(f"--output={output}")
    if resource_file:
        args.append(f"--resource-file={resource_file}")
    if resource_path:
        args.append(f"--resource-path={resource_path}")
    if textdomain:
        args.append(f"--textdomain={textdomain}")
    if locale:
        args.append(f"--locale={locale}")
    if locale_dir:
        args.append(f"--locale-dir={locale_dir}")
    if css:
        args.append(f"--css={css}")
    if scale:
        args.append(f"--scale={scale}")
    if dark:
        args.append(f"--dark")
    if libadwaita:
        args.append(f"--libadwaita")
    args.append(ui_file)
    env["WAYLAND_DISPLAY"] = SOCKET_2X if scale == 2 else SOCKET
    return subprocess.run(["uishooter"] + args, env=env).returncode


def shoot_application() -> None:
    """Shoot widgets for application metainfo."""
    print("Shooting application widgets…")
    total = 4
    for i in range(1, total + 1):
        widget = ApplicationWidget(i)
        print(f"[{i}/{total}] Shooting {widget.ui_file}…")
        exit_code = run_uishooter(
            ui_file=widget.ui_file,
            resource_path=RESOURCE_PATH,
            resource_file=RESOURCE_FILE,
            css=widget.css_file,
            output=widget.image_file,
            libadwaita=True)
        if exit_code != 0:
            raise RuntimeError(f"Error while shooting {widget.ui_file}.")
        write_license_file(widget.license_file)


def shoot_help() -> None:
    """Shoot widgets for the help pages."""
    print("Shooting help widgets…")
    with open(os.path.join("help", "LINGUAS")) as f:
        langcodes = [code for code in f.read().splitlines()
                     if code != "" and code[0] != "#"]
    langcodes.sort()
    total = len(langcodes) + 1
    for i, langcode in enumerate(["C"] + langcodes, 1):
        for widget in [
                HelpWidget("add-files-button", langcode, True),
                HelpWidget("add-folders-button", langcode, True),
                HelpWidget("clean-button", langcode, False),
                HelpWidget("metadata-example", langcode, True)]:
            print(f"[{i}/{total}|{langcode}] Shooting {widget.ui_file}…")
            exit_code = run_uishooter(
                ui_file=widget.ui_file,
                locale=locale_from_langcode(langcode),
                locale_dir=LOCALE_DIR,
                textdomain=TEXTDOMAIN,
                resource_path=RESOURCE_PATH,
                resource_file=RESOURCE_FILE,
                css=widget.css_file,
                output=widget.image_file,
                libadwaita=True)
            if exit_code != 0:
                raise RuntimeError(f"Error while shooting {widget.ui_file}.")
            write_license_file(widget.license_file)


def shoot_website() -> None:
    """Shoot widgets for website."""
    print("Shooting website widgets…")
    for scale in [1, 2]:
        image_suffix = f"-{scale}x" if scale > 1 else ""
        widget = Widget(
            os.path.join("screenshots", "website.ui"),
            os.path.join("website", f"app{image_suffix}.png"),
            os.path.join("screenshots", "website.css"))
        print(f"[1/1|{scale}x] Shooting {widget.ui_file}…")
        exit_code = run_uishooter(
            ui_file=widget.ui_file,
            resource_path=RESOURCE_PATH,
            resource_file=RESOURCE_FILE,
            css=widget.css_file,
            scale=scale,
            output=widget.image_file,
            libadwaita=True)
        if exit_code != 0:
            raise RuntimeError(f"Error while shooting {widget.ui_file}.")
        write_license_file(widget.license_file)


def write_license_file(path: str) -> None:
    """Write a license file at the given path.

    Args:
        path (str): The path of the license file.
    """
    with open(path, "w") as f:
        f.writelines([
            "SPDX-FileCopyrightText: Metadata Cleaner contributors\n"
            "SPDX-License-Identifier: CC-BY-SA-4.0"
        ])


def locale_from_langcode(langcode: str) -> str:
    """Get the locale string from a language code.

    Args:
        langcode (str): Language code.

    Returns:
        str: Locale string.
    """
    # Likely territory for "pt" is "BR", force "PT"
    if langcode == "pt":
        langcode = "pt-PT"
    try:
        language = Language.get(langcode)
        language = language.fill_likely_values()
        return f"{language.language}_{language.territory}.utf8"
    except LanguageTagError:
        return "C"


if __name__ == "__main__":
    weston = None
    weston_2x = None
    try:
        compile_translations()
        compile_resources()
        weston = start_weston(socket=SOCKET)
        weston_2x = start_weston(scale=2, socket=SOCKET_2X)
        shoot_application()
        shoot_help()
        shoot_website()
    finally:
        if weston:
            weston.terminate()
        if weston_2x:
            weston_2x.terminate()



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""View showing all the files."""

from gi.repository import Gio, GLib, GObject, Gtk

from metadatacleaner.modules.filestore import FileStore

from metadatacleaner.ui.cleanmetadatabutton import CleanMetadataButton
from metadatacleaner.ui.filerow import FileRow
from metadatacleaner.ui.settingsbutton import SettingsButton
from metadatacleaner.ui.statusindicator import StatusIndicator


@Gtk.Template(resource_path="/fr/romainvigier/MetadataCleaner/ui/FilesView.ui")
class FilesView(Gtk.Box):
    """View showing all the files."""

    __gtype_name__ = "FilesView"

    file_store = GObject.Property(type=FileStore, nick="file-store")

    _selection_model: Gtk.SingleSelection = Gtk.Template.Child()

    def __init__(self, *args, **kwargs) -> None:
        """View initialization."""
        super().__init__(*args, **kwargs)
        self._setup_actions()

    def _setup_actions(self) -> None:
        group = Gio.SimpleActionGroup.new()
        self.insert_action_group("files", group)

        def on_select(action: Gio.Action, parameters: GLib.Variant) -> None:
            self._selection_model.set_selected(parameters.get_uint32())
        select = Gio.SimpleAction.new("select", GLib.VariantType.new("u"))
        select.connect("activate", on_select)
        group.insert(select)

    @Gtk.Template.Callback()
    def _on_selection_changed(
            self,
            selection_model: Gtk.SelectionModel,
            position: int,
            n_items: int) -> None:
        number_of_items = len(selection_model.get_model())
        if selection_model.get_selected() >= number_of_items:
            return
        self.activate_action(
            "win.view-details",
            GLib.Variant.new_uint32(selection_model.get_selected()))

    def get_selected_file_index(self) -> int:
        """Get the index of the currently selected file.

        Returns:
            int: Index of the selected file.
        """
        return self._selection_model.get_selected()

    def clear_selected_file(self) -> None:
        """Unselect the currently selected file."""
        number_of_items = len(self._selection_model.get_model())
        if self._selection_model.get_selected() >= number_of_items:
            return
        self._selection_model.unselect_item(
            self._selection_model.get_selected())



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Button to clean metadata of all added files."""

from gi.repository import Gtk


@Gtk.Template(
    resource_path="/fr/romainvigier/MetadataCleaner/ui/CleanMetadataButton.ui"
)
class CleanMetadataButton(Gtk.Button):
    """Button to clean metadata of all added files."""

    __gtype_name__ = "CleanMetadataButton"
